const fs = require('fs-extra');
const Promise = require('bluebird');
const path = require('path');
const debug = require('ghost-ignition').debug('frontend:services:settings:ensure-settings');
const common = require('../../../server/lib/common');
const config = require('../../../server/config');

/**
 * Makes sure that all supported settings files are in the
 * `/content/settings` directory. If not, copy the default files
 * over.
 * @param {Array} knownSettings
 * @returns {Promise}
 * @description Reads the `/settings` folder of the content path and makes
 * sure that the associated yaml file for each setting exists. If it doesn't
 * copy the default yaml file over.
 */
module.exports = function ensureSettingsFiles(knownSettings) {
    const contentPath = config.getContentPath('settings');
    const defaultSettingsPath = config.get('paths').defaultSettings;

    return Promise.each(knownSettings, function (setting) {
        const fileName = `${setting}.yaml`;
        const defaultFileName = `default-${fileName}`;
        const filePath = path.join(contentPath, fileName);

        return fs.readFile(filePath, 'utf8')
            .catch({code: 'ENOENT'}, () => {
                // CASE: file doesn't exist, copy it from our defaults
                return fs.copy(
                    path.join(defaultSettingsPath, defaultFileName),
                    path.join(contentPath, fileName)
                ).then(() => {
                    debug(`'${defaultFileName}' copied to ${contentPath}.`);
                });
            }).catch((error) => {
                // CASE: we might have a permission error, as we can't access the directory
                throw new common.errors.GhostError({
                    message: common.i18n.t('errors.services.settings.ensureSettings', {path: contentPath}),
                    err: error,
                    context: error.path
                });
            });
    });
};
